@@ -3,10 +3,7 @@
 """
 Gives a multi-value dictionary object (MultiDict) plus several wrappers
 """
-import cgi
-import copy
-import sys
-
+import cgi, copy, sys, warnings, urllib
 from UserDict import DictMixin
 
 
@@ -56,12 +53,19 @@ class MultiDict(DictMixin):
         Create a dict from a cgi.FieldStorage instance
         """
         obj = cls()
+        if fs.type == 'multipart/form-data':
+            # cgi.FieldStorage parser for this ctype
+            # does not unquote the field names
+            fix_name = urllib.unquote
+        else:
+            fix_name = lambda n: n
         # fs.list can be None when there's nothing to parse
         for field in fs.list or ():
+            name = fix_name(field.name)
             if field.filename:
-                obj.add(field.name, field)
+                obj.add(name, field)
             else:
-                obj.add(field.name, field.value)
+                obj.add(name, field.value)
         return obj
 
     def __getitem__(self, key):
@@ -185,6 +189,18 @@ class MultiDict(DictMixin):
     def popitem(self):
         return self._items.pop()
 
+    def update(self, *args, **kw):
+        if args:
+            lst = args[0]
+            if len(lst) != len(dict(lst)):
+                # this does not catch the cases where we overwrite existing keys,
+                # but those would produce too many warning
+                msg = ("Behavior of MultiDict.update() has changed "
+                    "and overwrites duplicate keys. Consider using .extend()"
+                )
+                warnings.warn(msg, stacklevel=2)
+        DictMixin.update(self, *args, **kw)
+
     def extend(self, other=None, **kwargs):
         if other is None:
             pass
@@ -249,7 +265,7 @@ class UnicodeMultiDict(DictMixin):
     is enabled.
 
     """
-    def __init__(self, multi=None, encoding=None, errors='strict',
+    def __init__(self, multi, encoding=None, errors='strict',
                  decode_keys=False):
         self.multi = multi
         if encoding is None:
@@ -448,8 +464,8 @@ class TrackableMultiDict(MultiDict):
         result = MultiDict.popitem(self)
         self.tracker(self)
         return result
-    def update(self, other=None, **kwargs):
-        MultiDict.update(self, other, **kwargs)
+    def update(self, *args, **kwargs):
+        MultiDict.update(self, *args, **kwargs)
         self.tracker(self)
     def __repr__(self):
         items = ', '.join(['(%r, %r)' % v for v in self.iteritems()])
